#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

import os
from importlib.metadata import PackageNotFoundError, version

import breezy
from breezy import urlutils

from . import __revision__, __version__
from .util import html_format
from .zptsupport import zpt

templatefunctions = {}


def templatefunc(func):
    templatefunctions[func.__name__] = func
    return func


_base = os.path.dirname(__file__)


def _pt(name):
    return zpt(os.path.join(_base, "templates", name + ".pt"))


templatefunctions["macros"] = _pt("macros").macros
templatefunctions["breadcrumbs"] = _pt("breadcrumbs").macros


@templatefunc
def file_change_summary(
    url, entry, file_changes, style="normal", currently_showing=None
):
    if style == "fragment":

        def file_link(filename):
            if currently_showing and filename == currently_showing:
                return html_format(
                    '<b><a href="#%s">%s</a></b>', urlutils.quote(filename), filename
                )
            else:
                return revision_link(
                    url, entry.revno, filename, "#" + urlutils.quote(filename)
                )
    else:

        def file_link(filename):
            return html_format(
                '<a href="%s%s" title="View changes to %s in revision %s">%s</a>',
                url(["/revision", entry.revno]),
                "#" + urlutils.quote(filename),
                filename,
                entry.revno,
                filename,
            )

    return _pt("revisionfilechanges").expand(
        entry=entry, file_changes=file_changes, file_link=file_link, **templatefunctions
    )


@templatefunc
def revisioninfo(
    url, branch, entry, file_changes=None, currently_showing=None, merged_in=None
):
    from . import util

    return _pt("revisioninfo").expand(
        url=url,
        change=entry,
        branch=branch,
        util=util,
        file_changes=file_changes,
        currently_showing=currently_showing,
        merged_in=merged_in,
        **templatefunctions,
    )


@templatefunc
def branchinfo(branch):
    if branch.served_url is not None:
        return _pt("branchinfo").expand(branch=branch, **templatefunctions)
    else:
        return ""


@templatefunc
def collapse_button(group, name, branch, normal="block"):
    return _pt("collapse-button").expand(
        group=group, name=name, normal=normal, branch=branch, **templatefunctions
    )


@templatefunc
def collapse_all_button(group, branch, normal="block"):
    return _pt("collapse-all-button").expand(
        group=group, normal=normal, branch=branch, **templatefunctions
    )


@templatefunc
def revno_with_nick(entry):
    if entry.branch_nick:
        extra = " " + entry.branch_nick
    else:
        extra = ""
    return "(%s%s)" % (entry.revno, extra)


@templatefunc
def search_box(branch, navigation):
    return _pt("search-box").expand(
        branch=branch, navigation=navigation, **templatefunctions
    )


@templatefunc
def feed_link(branch, url):
    return _pt("feed-link").expand(branch=branch, url=url, **templatefunctions)


@templatefunc
def menu(branch, url, fileview_active=False):
    return _pt("menu").expand(
        branch=branch, url=url, fileview_active=fileview_active, **templatefunctions
    )


@templatefunc
def view_link(url, revno, path):
    return html_format(
        '<a href="%s" title="Annotate %s">%s</a>',
        url(["/view", revno, path]),
        path,
        path,
    )


@templatefunc
def revision_link(url, revno, path, frag=""):
    return html_format(
        '<a href="%s%s" title="View changes to %s in revision %s">%s</a>',
        url(["/revision", revno, path]),
        frag,
        path,
        revno,
        path,
    )


@templatefunc
def loggerhead_version():
    return __version__


@templatefunc
def loggerhead_revision():
    return __revision__


_cached_generator_string = None


@templatefunc
def generator_string():
    global _cached_generator_string
    if _cached_generator_string is None:
        versions = []

        # TODO: Errors -- e.g. from a missing/invalid __version__ attribute, or
        # ValueError accessing Distribution.version -- should be non-fatal.

        versions.append(("Loggerhead", __version__))

        import sys

        python_version = breezy._format_version_tuple(sys.version_info)
        versions.append(("Python", python_version))

        versions.append(("Breezy", breezy.__version__))

        versions.append(("Paste", version("Paste")))

        try:
            paste_deploy_version = version("PasteDeploy")
        except PackageNotFoundError:
            pass
        else:
            versions.append(("PasteDeploy", paste_deploy_version))

        versions.append(("Chameleon", version("Chameleon")))

        try:
            import pygments
        except ImportError:
            pass
        else:
            versions.append(("Pygments", pygments.__version__))

        try:
            from breezy.plugins import search
        except ImportError:
            pass
        else:
            bzr_search_version = breezy._format_version_tuple(search.version_info)
            versions.append(("bzr-search", bzr_search_version))

        # TODO: On old Python versions, elementtree may be used.

        try:
            dozer_version = version("Dozer")
        except PackageNotFoundError:
            pass
        else:
            versions.append(("Dozer", dozer_version))

        version_strings = ("%s/%s" % t for t in versions)
        _cached_generator_string = " ".join(version_strings)
    return _cached_generator_string



# Copyright (C) 2011 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""Direct tests of the loggerhead/history.py module"""

from datetime import datetime

from breezy import tag, tests
from breezy.foreign import ForeignRevision, ForeignVcs, VcsMapping

from .. import history as _mod_history


class TestCaseWithExamples(tests.TestCaseWithMemoryTransport):
    def make_linear_ancestry(self):
        # Time goes up
        # rev-3
        #  |
        # rev-2
        #  |
        # rev-1
        builder = self.make_branch_builder("branch")
        builder.start_series()
        rev1 = builder.build_snapshot(
            None, [("add", ("", b"root-id", "directory", None))]
        )
        rev2 = builder.build_snapshot([rev1], [])
        rev3 = builder.build_snapshot([rev2], [])
        builder.finish_series()
        b = builder.get_branch()
        self.addCleanup(b.lock_read().unlock)
        return _mod_history.History(b, {}), [rev1, rev2, rev3]

    def make_long_linear_ancestry(self):
        builder = self.make_branch_builder("branch")
        revs = []
        builder.start_series()
        revs.append(
            builder.build_snapshot(None, [("add", ("", b"root-id", "directory", None))])
        )
        for r in "BCDEFGHIJKLMNOPQRSTUVWXYZ":
            revs.append(builder.build_snapshot(None, []))
        builder.finish_series()
        b = builder.get_branch()
        self.addCleanup(b.lock_read().unlock)
        return _mod_history.History(b, {}), revs

    def make_merged_ancestry(self):
        # Time goes up
        # rev-3
        #  |  \
        #  |  rev-2
        #  |  /
        # rev-1
        builder = self.make_branch_builder("branch")
        builder.start_series()
        rev1 = builder.build_snapshot(
            None, [("add", ("", b"root-id", "directory", None))]
        )
        rev2 = builder.build_snapshot([rev1], [])
        rev3 = builder.build_snapshot([rev1, rev2], [])
        builder.finish_series()
        b = builder.get_branch()
        self.addCleanup(b.lock_read().unlock)
        return _mod_history.History(b, {}), [rev1, rev2, rev3]

    def make_deep_merged_ancestry(self):
        # Time goes up
        # F
        # |\
        # | E
        # | |\
        # | | D
        # | |/
        # B C
        # |/
        # A
        builder = self.make_branch_builder("branch")
        builder.start_series()
        rev_a = builder.build_snapshot(
            None, [("add", ("", b"root-id", "directory", None))]
        )
        rev_b = builder.build_snapshot([rev_a], [])
        rev_c = builder.build_snapshot([rev_a], [])
        rev_d = builder.build_snapshot([rev_c], [])
        rev_e = builder.build_snapshot([rev_c, rev_d], [])
        rev_f = builder.build_snapshot([rev_b, rev_e], [])
        builder.finish_series()
        b = builder.get_branch()
        self.addCleanup(b.lock_read().unlock)
        return (_mod_history.History(b, {}), [rev_a, rev_b, rev_c, rev_d, rev_e, rev_f])

    def assertRevidsFrom(self, expected, history, search_revs, tip_rev):
        self.assertEqual(expected, list(history.get_revids_from(search_revs, tip_rev)))


class _DictProxy:
    def __init__(self, d):
        self._d = d
        self._accessed = set()
        self.__setitem__ = d.__setitem__

    def __getitem__(self, name):
        self._accessed.add(name)
        return self._d[name]

    def __len__(self):
        return len(self._d)


def track_rev_info_accesses(h):
    """Track __getitem__ access to History._rev_info,

    :return: set of items accessed
    """
    h._rev_info = _DictProxy(h._rev_info)
    return h._rev_info._accessed


class TestHistoryGetRevidsFrom(TestCaseWithExamples):
    def test_get_revids_from_simple_mainline(self):
        history, revs = self.make_linear_ancestry()
        self.assertRevidsFrom(list(reversed(revs)), history, None, revs[2])

    def test_get_revids_from_merged_mainline(self):
        history, revs = self.make_merged_ancestry()
        self.assertRevidsFrom([revs[2], revs[0]], history, None, revs[2])

    def test_get_revids_given_one_rev(self):
        history, revs = self.make_merged_ancestry()
        # rev-3 was the first mainline revision to see rev-2.
        self.assertRevidsFrom([revs[2]], history, [revs[1]], revs[2])

    def test_get_revids_deep_ancestry(self):
        history, revs = self.make_deep_merged_ancestry()
        self.assertRevidsFrom([revs[-1]], history, [revs[-1]], revs[-1])
        self.assertRevidsFrom([revs[-1]], history, [revs[-2]], revs[-1])
        self.assertRevidsFrom([revs[-1]], history, [revs[-3]], revs[-1])
        self.assertRevidsFrom([revs[-1]], history, [revs[-4]], revs[-1])
        self.assertRevidsFrom([revs[1]], history, [revs[-5]], revs[-1])
        self.assertRevidsFrom([revs[0]], history, [revs[-6]], revs[-1])

    def test_get_revids_doesnt_over_produce_simple_mainline(self):
        # get_revids_from shouldn't walk the whole ancestry just to get the
        # answers for the first few revisions.
        history, revs = self.make_long_linear_ancestry()
        accessed = track_rev_info_accesses(history)
        result = history.get_revids_from(None, revs[-1])
        self.assertEqual(set(), accessed)
        self.assertEqual(revs[-1], next(result))
        # We already know revs[-1] because we passed it in.
        self.assertEqual(set(), accessed)
        self.assertEqual(revs[-2], next(result))
        self.assertEqual({history._rev_indices[revs[-1]]}, accessed)

    def test_get_revids_doesnt_over_produce_for_merges(self):
        # get_revids_from shouldn't walk the whole ancestry just to get the
        # answers for the first few revisions.
        history, revs = self.make_long_linear_ancestry()
        accessed = track_rev_info_accesses(history)
        result = history.get_revids_from([revs[-3], revs[-5]], revs[-1])
        self.assertEqual(set(), accessed)
        self.assertEqual(revs[-3], next(result))
        # We access 'W' because we are checking that W wasn't merged into X.
        # The important bit is that we aren't getting the whole ancestry.
        self.assertEqual(
            {history._rev_indices[x] for x in list(reversed(revs))[:4]}, accessed
        )
        self.assertEqual(revs[-5], next(result))
        self.assertEqual(
            {history._rev_indices[x] for x in list(reversed(revs))[:6]}, accessed
        )
        self.assertRaises(StopIteration, next, result)
        self.assertEqual(
            {history._rev_indices[x] for x in list(reversed(revs))[:6]}, accessed
        )


class TestHistoryChangeFromRevision(tests.TestCaseWithTransport):
    def make_single_commit(self):
        tree = self.make_branch_and_tree("test")
        rev_id = tree.commit(
            "Commit Message",
            timestamp=1299838474.317,
            timezone=3600,
            committer="Joe Example <joe@example.com>",
            revprops={},
        )
        self.addCleanup(tree.branch.lock_write().unlock)
        rev = tree.branch.repository.get_revision(rev_id)
        history = _mod_history.History(tree.branch, {})
        return history, rev

    def test_simple(self):
        history, rev = self.make_single_commit()
        change = history._change_from_revision(rev)
        self.assertEqual(rev.revision_id, change.revid)
        self.assertEqual(datetime.fromtimestamp(1299838474.317), change.date)
        self.assertEqual(datetime.utcfromtimestamp(1299838474.317), change.utc_date)
        self.assertEqual(["Joe Example <joe@example.com>"], change.authors)
        self.assertEqual("test", change.branch_nick)
        self.assertEqual("Commit Message", change.short_comment)
        self.assertEqual("Commit Message", change.comment)
        self.assertEqual(["Commit&nbsp;Message"], change.comment_clean)
        self.assertEqual([], change.parents)
        self.assertEqual([], change.bugs)
        self.assertEqual(None, change.tags)

    def test_tags(self):
        history, rev = self.make_single_commit()
        b = history._branch
        b.tags.set_tag("tag-1", rev.revision_id)
        b.tags.set_tag("tag-2", rev.revision_id)
        b.tags.set_tag("Tag-10", rev.revision_id)
        change = history._change_from_revision(rev)
        # If available, tags are 'naturally' sorted. (sorting numbers in order,
        # and ignoring case, etc.)
        if getattr(tag, "sort_natural", None) is not None:
            self.assertEqual("tag-1, tag-2, Tag-10", change.tags)
        else:
            self.assertEqual("Tag-10, tag-1, tag-2", change.tags)

    def test_committer_vs_authors(self):
        tree = self.make_branch_and_tree("test")
        rev_id = tree.commit(
            "Commit Message",
            timestamp=1299838474.317,
            timezone=3600,
            committer="Joe Example <joe@example.com>",
            revprops={
                "authors": "A Author <aauthor@example.com>\n"
                "B Author <bauthor@example.com>"
            },
        )
        self.addCleanup(tree.branch.lock_write().unlock)
        rev = tree.branch.repository.get_revision(rev_id)
        history = _mod_history.History(tree.branch, {})
        change = history._change_from_revision(rev)
        self.assertEqual("Joe Example <joe@example.com>", change.committer)
        self.assertEqual(
            ["A Author <aauthor@example.com>", "B Author <bauthor@example.com>"],
            change.authors,
        )


class TestHistory_IterateSufficiently(tests.TestCase):
    def assertIterate(self, expected, iterable, stop_at, extra_rev_count):
        self.assertEqual(
            expected,
            _mod_history.History._iterate_sufficiently(
                iterable, stop_at, extra_rev_count
            ),
        )

    def test_iter_no_extra(self):
        lst = list("abcdefghijklmnopqrstuvwxyz")
        self.assertIterate(["a", "b", "c"], iter(lst), "c", 0)
        self.assertIterate(["a", "b", "c", "d"], iter(lst), "d", 0)

    def test_iter_not_found(self):
        # If the key in question isn't found, we just exhaust the list
        lst = list("abcdefghijklmnopqrstuvwxyz")
        self.assertIterate(lst, iter(lst), "not-there", 0)

    def test_iter_with_extra(self):
        lst = list("abcdefghijklmnopqrstuvwxyz")
        self.assertIterate(["a", "b", "c"], iter(lst), "b", 1)
        self.assertIterate(["a", "b", "c", "d", "e"], iter(lst), "c", 2)

    def test_iter_with_too_many_extra(self):
        lst = list("abcdefghijklmnopqrstuvwxyz")
        self.assertIterate(lst, iter(lst), "y", 10)
        self.assertIterate(lst, iter(lst), "z", 10)

    def test_iter_with_extra_None(self):
        lst = list("abcdefghijklmnopqrstuvwxyz")
        self.assertIterate(lst, iter(lst), "z", None)


class TestHistoryGetView(TestCaseWithExamples):
    def test_get_view_limited_history(self):
        # get_view should only load enough history to serve the result, not all
        # history.
        history, revs = self.make_long_linear_ancestry()
        accessed = track_rev_info_accesses(history)
        revid, start_revid, revid_list = history.get_view(
            revs[-1], revs[-1], None, extra_rev_count=5
        )
        self.assertEqual(list(reversed(revs))[:6], revid_list)
        self.assertEqual(revs[-1], revid)
        self.assertEqual(revs[-1], start_revid)
        self.assertEqual(
            {history._rev_indices[x] for x in list(reversed(revs))[:6]}, accessed
        )


class TestHistoryGetChangedUncached(TestCaseWithExamples):
    def test_native(self):
        history, revs = self.make_linear_ancestry()
        changes = history.get_changes_uncached([revs[0], revs[1]])
        self.assertEqual(2, len(changes))
        self.assertEqual(revs[0], changes[0].revid)
        self.assertEqual(revs[1], changes[1].revid)
        self.assertIs(None, getattr(changes[0], "foreign_vcs", None))
        self.assertIs(None, getattr(changes[0], "foreign_revid", None))

    def test_foreign(self):
        # Test with a mocked foreign revision, as it's not possible
        # to rely on any foreign plugins being installed.
        history, revs = self.make_linear_ancestry()
        foreign_vcs = ForeignVcs(None, "vcs")
        foreign_vcs.show_foreign_revid = repr
        foreign_rev = ForeignRevision(
            ("uuid", 1234),
            VcsMapping(foreign_vcs),
            "revid-in-bzr",
            message="message",
            timestamp=234423423.3,
        )
        change = history._change_from_revision(foreign_rev)
        self.assertEqual("revid-in-bzr", change.revid)
        self.assertEqual("('uuid', 1234)", change.foreign_revid)
        self.assertEqual("vcs", change.foreign_vcs)



"""Profiling middleware for Paste."""

import threading

from breezy.lsprof import profile


class LSProfMiddleware(object):
    """Paste middleware for profiling with lsprof."""

    def __init__(self, app, global_conf=None):
        self.app = app
        self.lock = threading.Lock()
        self.__count = 0

    def __run_app(self, environ, start_response):
        app_iter = self.app(environ, start_response)
        try:
            return list(app_iter)
        finally:
            if getattr(app_iter, "close", None):
                app_iter.close()

    def __call__(self, environ, start_response):
        """Run a request."""
        self.lock.acquire()
        try:
            ret, stats = profile(self.__run_app, environ, start_response)
            self.__count += 1
            stats.save("%d-stats.callgrind" % (self.__count,), format="callgrind")
            return ret
        finally:
            self.lock.release()



# Copyright (C) 2007, 2008, 2009, 2011 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import absolute_import

from fixtures import Fixture


class SampleBranch(Fixture):
    def __init__(self, testcase):
        # Must be a bzr TestCase to hook into branch creation, unfortunately.
        self.testcase = testcase

    def setUp(self):
        Fixture.setUp(self)

        self.tree = self.testcase.make_branch_and_tree(".")

        self.filecontents = "some\nmultiline\ndata\nwith<htmlspecialchars\n"
        filenames = ["myfilename", "anotherfile<", "folder/", "folder/myfilename"]
        self.testcase.build_tree_contents(
            (filename, self.filecontents) for filename in filenames
        )
        self.tree.add(filenames)
        self.path = "myfilename"
        self.msg = "a very exciting commit message <"
        self.revid = self.tree.commit(message=self.msg, rev_id=b"rev-1")



"""WSGI applications for serving Bazaar branches."""

import os

from paste import fileapp, urlparser

from ..util import convert_file_errors

static = os.path.join(os.path.dirname(os.path.dirname(__file__)), "static")

# Static things can be cached for half a day, we could probably make this
# longer, except for just before rollout times.
static_app = urlparser.make_static(None, static, cache_max_age=12 * 60 * 60)

favicon_app = convert_file_errors(
    fileapp.FileApp(os.path.join(static, "images", "favicon.ico"))
)

robots_app = convert_file_errors(fileapp.FileApp(os.path.join(static, "robots.txt")))


def health_app(environ, start_response):
    start_response("200 OK", [])
    yield b"ok"



#
# Copyright (C) 2008, 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA

"""Search for branches underneath a directory and serve them all."""

import logging
import os
import sys

from breezy.location import location_to_url
from breezy.plugin import load_plugins
from paste import httpserver
from paste.httpexceptions import HTTPExceptionHandler, HTTPInternalServerError
from paste.translogger import TransLogger

from . import __version__
from .apps.error import ErrorHandlerApp
from .apps.transport import BranchesFromTransportRoot, UserBranchesFromTransportRoot
from .config import LoggerheadConfig
from .util import Reloader


def get_config_and_base(args):
    config = LoggerheadConfig(args)

    if config.get_option("show_version"):
        print("loggerhead %s" % (__version__,))
        sys.exit(0)

    if config.arg_count > 1:
        config.print_help()
        sys.exit(1)
    elif config.arg_count == 1:
        base = config.get_arg(0)
    else:
        base = "."

    base = location_to_url(base)

    if not config.get_option("allow_writes"):
        base = "readonly+" + base

    return config, base


def setup_logging(config, init_logging=True, log_file=None):
    log_level = config.get_log_level()
    if init_logging:
        logging.basicConfig()
        if log_level is not None:
            logging.getLogger("").setLevel(log_level)
    logger = logging.getLogger("loggerhead")
    if log_level is not None:
        logger.setLevel(log_level)
    if log_file is not None:
        handler = logging.StreamHandler(log_file)
    else:
        if config.get_option("log_folder"):
            logfile_path = os.path.join(
                config.get_option("log_folder"), "loggerhead-serve.log"
            )
        else:
            logfile_path = "loggerhead-serve.log"
        handler = logging.FileHandler(logfile_path, "a")
        formatter = logging.Formatter(
            "%(asctime)s %(levelname)-8s %(name)s: %(message)s"
        )
        handler.setFormatter(formatter)
    # We set the handler to accept all messages, the *logger* won't emit them
    # if it is configured to suppress it
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    return logger


def make_app_for_config_and_base(config, base):
    if config.get_option("trunk_dir") and not config.get_option("user_dirs"):
        print("--trunk-dir is only valid with --user-dirs")
        sys.exit(1)

    if config.get_option("reload"):
        if Reloader.is_installed():
            Reloader.install()
        else:
            return Reloader.restart_with_reloader()

    if config.get_option("user_dirs"):
        if not config.get_option("trunk_dir"):
            print("You didn't specify a directory for the trunk directories.")
            sys.exit(1)
        app = UserBranchesFromTransportRoot(base, config)
    else:
        app = BranchesFromTransportRoot(base, config)

    setup_logging(config)

    if config.get_option("profile"):
        from loggerhead.middleware.profile import LSProfMiddleware

        app = LSProfMiddleware(app)
    if config.get_option("memory_profile"):
        from dozer import Dozer

        app = Dozer(app)

    if not config.get_option("user_prefix"):
        prefix = "/"
    else:
        prefix = config.get_option("user_prefix")
        if not prefix.startswith("/"):
            prefix = "/" + prefix

    try:
        from paste.deploy.config import PrefixMiddleware
    except ImportError:
        cant_proxy_correctly_message = (
            "Unsupported configuration: PasteDeploy not available, but "
            "loggerhead appears to be behind a proxy."
        )

        def check_not_proxied(app):
            def wrapped(environ, start_response):
                if "HTTP_X_FORWARDED_SERVER" in environ:
                    exc = HTTPInternalServerError()
                    exc.explanation = cant_proxy_correctly_message
                    raise exc
                return app(environ, start_response)

            return wrapped

        logging.warning(
            "PasteDeploy not available; unable to support "
            "access through a reverse proxy."
        )
        app = check_not_proxied(app)
    else:
        app = PrefixMiddleware(app, prefix=prefix)

    app = HTTPExceptionHandler(app)
    app = ErrorHandlerApp(app)
    app = TransLogger(app, logger=logging.getLogger("loggerhead"))

    return app


def main(args):
    load_plugins()

    config, base = get_config_and_base(args)

    app = make_app_for_config_and_base(config, base)

    if not config.get_option("user_port"):
        port = "8080"
    else:
        port = config.get_option("user_port")

    if not config.get_option("user_host"):
        host = "0.0.0.0"
    else:
        host = config.get_option("user_host")

    if not config.get_option("protocol"):
        protocol = "http"
    else:
        protocol = config.get_option("protocol")

    if protocol == "http":
        httpserver.serve(app, host=host, port=port)
    else:
        if protocol == "fcgi":
            from flup.server.fcgi import WSGIServer
        elif protocol == "scgi":
            from flup.server.scgi import WSGIServer
        elif protocol == "ajp":
            from flup.server.ajp import WSGIServer
        else:
            print("Unknown protocol: %s." % (protocol))
            sys.exit(1)
        WSGIServer(app, bindAddress=(host, int(port))).run()


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Copyright 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Tests for the HeadMiddleware app."""

from io import BytesIO

from breezy import tests

from ..apps import http_head

content = [
    b"<html>",
    b"<head><title>Listed</title></head>",
    b"<body>Content</body>",
    b"</html>",
]
headers = {"X-Ignored-Header": "Value"}


def yielding_app(environ, start_response):
    start_response("200 OK", headers)
    yield from content


def list_app(environ, start_response):
    start_response("200 OK", headers)
    return content


def writer_app(environ, start_response):
    writer = start_response("200 OK", headers)
    for chunk in content:
        writer(chunk)
    return []


class TestHeadMiddleware(tests.TestCase):
    def _trap_start_response(self, status, response_headers, exc_info=None):
        self._write_buffer = BytesIO()
        self._start_response_passed = (status, response_headers, exc_info)
        return self._write_buffer.write

    def _consume_app(self, app, request_method):
        environ = {"REQUEST_METHOD": request_method}
        value = list(app(environ, self._trap_start_response))
        self._write_buffer.writelines(value)

    def _verify_get_passthrough(self, app):
        app = http_head.HeadMiddleware(app)
        self._consume_app(app, "GET")
        self.assertEqual(("200 OK", headers, None), self._start_response_passed)
        self.assertEqualDiff(b"".join(content), self._write_buffer.getvalue())

    def _verify_head_no_body(self, app):
        app = http_head.HeadMiddleware(app)
        self._consume_app(app, "HEAD")
        self.assertEqual(("200 OK", headers, None), self._start_response_passed)
        self.assertEqualDiff(b"", self._write_buffer.getvalue())

    def test_get_passthrough_yielding(self):
        self._verify_get_passthrough(yielding_app)

    def test_head_passthrough_yielding(self):
        self._verify_head_no_body(yielding_app)

    def test_get_passthrough_list(self):
        self._verify_get_passthrough(list_app)

    def test_head_passthrough_list(self):
        self._verify_head_no_body(list_app)

    def test_get_passthrough_writer(self):
        self._verify_get_passthrough(writer_app)

    def test_head_passthrough_writer(self):
        self._verify_head_no_body(writer_app)



# Copyright (C) 2008  Canonical Ltd.
#                     (Authored by Martin Albisetti <argentina@gmail.com>)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

import datetime
import logging
import stat

from breezy import branch, errors, urlutils

try:
    from breezy.transport import NoSuchFile
except ImportError:
    from breezy.errors import NoSuchFile

from .. import util
from ..controllers import TemplatedBranchView


class DirEntry(object):
    def __init__(self, dirname, parity, branch, transport):
        self.dirname = urlutils.unquote(dirname)
        self.parity = parity
        self.branch = branch
        if branch is None:
            self.last_revision = None
            try:
                self.last_change_time = datetime.datetime.utcfromtimestamp(
                    transport.stat(self.dirname).st_mtime
                )
            except Exception:
                self.last_change_time = None
        else:
            # If a branch is empty, bzr raises an exception when trying this
            try:
                self.last_revision = branch.repository.get_revision(
                    branch.last_revision()
                )
                self.last_change_time = datetime.datetime.utcfromtimestamp(
                    self.last_revision.timestamp
                )
            except errors.NoSuchRevision:
                self.last_revision = None
                self.last_change_time = None


class DirectoryUI(TemplatedBranchView):
    """ """

    template_name = "directory"

    def __init__(self, static_url_base, transport, name):
        class _branch(object):
            context_url = 1

            @staticmethod
            def static_url(path):
                return self._static_url_base + path

        self._branch = _branch
        self._history_callable = lambda: None
        self._name = name
        self._static_url_base = static_url_base
        self.transport = transport
        self.log = logging.getLogger("")

    def get_values(self, path, kwargs, response):
        listing = [d for d in self.transport.list_dir(".") if not d.startswith(".")]
        listing.sort(key=lambda x: x.lower())
        dirs = []
        parity = 0
        for d in listing:
            try:
                b = branch.Branch.open_from_transport(self.transport.clone(d))
            except BaseException:
                # TODO(jelmer): don't catch all exceptions here
                try:
                    if not stat.S_ISDIR(self.transport.stat(d).st_mode):
                        continue
                except NoSuchFile:
                    continue
                b = None
            else:
                if not b.get_config().get_user_option_as_bool(
                    "http_serve", default=True
                ):
                    continue
            dirs.append(DirEntry(d, parity, b, self.transport))
            parity = 1 - parity
        # Create breadcrumb trail
        directory_breadcrumbs = util.directory_breadcrumbs(
            self._name, False, "directory"
        )
        return {
            "dirs": dirs,
            "name": self._name,
            "directory_breadcrumbs": directory_breadcrumbs,
        }



# Copyright (C) 2007, 2008, 2009, 2011 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


import json
import logging
import re
from html import escape
from io import BytesIO

from breezy import config
from breezy.tests import TestCaseWithTransport
from paste.fixture import TestApp
from paste.httpexceptions import HTTPExceptionHandler, HTTPMovedPermanently

from ..apps.branch import BranchWSGIApp
from ..apps.http_head import HeadMiddleware
from .fixtures import SampleBranch


class BasicTests(TestCaseWithTransport):
    def setUp(self):
        TestCaseWithTransport.setUp(self)
        logging.basicConfig(level=logging.ERROR)
        logging.getLogger("bzr").setLevel(logging.CRITICAL)

    def createBranch(self):
        self.tree = self.make_branch_and_tree(".")

    def setUpLoggerhead(self, **kw):
        branch_app = BranchWSGIApp(self.tree.branch, "", **kw).app
        return TestApp(HTTPExceptionHandler(branch_app))

    def assertOkJsonResponse(self, app, env):
        start, content = consume_app(app, env)
        self.assertEqual("200 OK", start[0])
        self.assertEqual("application/json", dict(start[1])["Content-Type"])
        self.assertEqual(None, start[2])
        json.loads(content.decode("UTF-8"))

    def make_branch_app(self, branch, **kw):
        branch_app = BranchWSGIApp(branch, friendly_name="friendly-name", **kw)
        branch_app._environ = {
            "wsgi.url_scheme": "",
            "SERVER_NAME": "",
            "SERVER_PORT": "80",
        }
        branch_app._url_base = ""
        return branch_app


class TestWithSimpleTree(BasicTests):
    def setUp(self):
        BasicTests.setUp(self)
        self.sample_branch_fixture = SampleBranch(self)

        # XXX: This could be cleaned up more... -- mbp 2011-11-25
        self.useFixture(self.sample_branch_fixture)
        self.tree = self.sample_branch_fixture.tree
        self.path = self.sample_branch_fixture.path
        self.filecontents = self.sample_branch_fixture.filecontents
        self.msg = self.sample_branch_fixture.msg

    def test_public_private(self):
        app = self.make_branch_app(self.tree.branch, private=True)
        self.assertEqual(app.public_private_css(), "private")
        app = self.make_branch_app(self.tree.branch)
        self.assertEqual(app.public_private_css(), "public")

    def test_changes(self):
        app = self.setUpLoggerhead()
        res = app.get("/changes")
        res.mustcontain(escape(self.msg))

    def test_changes_for_file(self):
        app = self.setUpLoggerhead()
        res = app.get("/changes?filter_path=%s" % self.path)
        res.mustcontain(escape(self.msg))

    def test_changes_branch_from(self):
        app = self.setUpLoggerhead(served_url="lp:loggerhead")
        res = app.get("/changes")
        self.assertIn("To get this branch, use:", res)
        self.assertIn("lp:loggerhead", res)

    def test_changes_search(self):
        app = self.setUpLoggerhead()
        res = app.get("/changes", params={"q": "foo"})
        res.mustcontain("Sorry, no results found for your search.")

    def test_annotate(self):
        app = self.setUpLoggerhead()
        res = app.get("/annotate/1/%s" % self.path, params={})
        # If pygments is installed, it inserts <span class="pyg" content into
        # the output, to trigger highlighting. And it specifically highlights
        # the &lt; that we are interested in seeing in the output.
        # Without pygments we have a simple: 'with&lt;htmlspecialchars'
        # With it, we have
        # '<span class='pyg-n'>with</span><span class='pyg-o'>&lt;</span>'
        # '<span class='pyg-n'>htmlspecialchars</span>
        # So we pre-filter the body, to make sure remove spans of that type.
        body_no_span = re.sub(b'<span class="pyg-.">', b"", res.body)
        body_no_span = body_no_span.replace(b"</span>", b"")
        for line in self.filecontents.splitlines():
            escaped = escape(line).encode("utf-8")
            self.assertIn(
                escaped,
                body_no_span,
                "did not find {!r} in {!r}".format(escaped, body_no_span),
            )

    def test_inventory(self):
        app = self.setUpLoggerhead()
        res = app.get("/files")
        res.mustcontain("myfilename")
        res = app.get("/files/")
        res.mustcontain("myfilename")
        res = app.get("/files/1")
        res.mustcontain("myfilename")
        res = app.get("/files/1/")
        res.mustcontain("myfilename")

    def test_inventory_bad_rev_404(self):
        app = self.setUpLoggerhead()
        app.get("/files/200", status=404)
        app.get("/files/invalid-revid", status=404)

    def test_inventory_bad_path_404(self):
        app = self.setUpLoggerhead()
        app.get("/files/1/hooha", status=404)

    def test_revision(self):
        app = self.setUpLoggerhead()
        res = app.get("/revision/1")
        res.mustcontain(no=["anotherfile<"])
        res.mustcontain("anotherfile&lt;")
        res.mustcontain("myfilename")


class TestEmptyBranch(BasicTests):
    """Test that an empty branch doesn't break"""

    def setUp(self):
        BasicTests.setUp(self)
        self.createBranch()

    def test_changes(self):
        app = self.setUpLoggerhead()
        res = app.get("/changes")
        res.mustcontain("No revisions!")

    def test_inventory(self):
        app = self.setUpLoggerhead()
        res = app.get("/files")
        res.mustcontain("No revisions!")


class TestHiddenBranch(BasicTests):
    """
    Test that hidden branches aren't shown
    FIXME: not tested that it doesn't show up on listings
    """

    def setUp(self):
        BasicTests.setUp(self)
        self.createBranch()
        try:
            locations = config.locations_config_filename()
        except AttributeError:
            from breezy import bedding

            locations = bedding.locations_config_path()
            ensure_config_dir_exists = bedding.ensure_config_dir_exists
        else:
            ensure_config_dir_exists = config.ensure_config_dir_exists
        ensure_config_dir_exists()
        with open(locations, "w") as f:
            f.write("[{}]\nhttp_serve = False".format(self.tree.branch.base))

    def test_no_access(self):
        app = self.setUpLoggerhead()
        app.get("/changes", status=404)


class TestControllerRedirects(BasicTests):
    """
    Test that a file under /files redirects to /view,
    and a directory under /view redirects to /files.
    """

    def setUp(self):
        BasicTests.setUp(self)
        self.createBranch()
        self.build_tree(("file", "folder/", "folder/file"))
        self.tree.smart_add([])
        self.tree.commit("")

    def test_view_folder(self):
        app = TestApp(BranchWSGIApp(self.tree.branch, "").app)

        e = self.assertRaises(HTTPMovedPermanently, app.get, "/view/head:/folder")
        self.assertEqual(e.location(), "/files/head:/folder")

    def test_files_file(self):
        app = TestApp(BranchWSGIApp(self.tree.branch, "").app)

        e = self.assertRaises(HTTPMovedPermanently, app.get, "/files/head:/folder/file")
        self.assertEqual(e.location(), "/view/head:/folder/file")
        e = self.assertRaises(HTTPMovedPermanently, app.get, "/files/head:/file")
        self.assertEqual(e.location(), "/view/head:/file")


class TestHeadMiddleware(BasicTests):
    def setUp(self):
        BasicTests.setUp(self)
        self.createBranch()
        self.msg = "trivial commit message"
        self.revid = self.tree.commit(message=self.msg)

    def setUpLoggerhead(self, **kw):
        branch_app = BranchWSGIApp(self.tree.branch, "", **kw).app
        return TestApp(HTTPExceptionHandler(HeadMiddleware(branch_app)))

    def test_get(self):
        app = self.setUpLoggerhead()
        res = app.get("/changes")
        res.mustcontain(self.msg)
        self.assertEqual("text/html", res.header("Content-Type"))

    def test_head(self):
        app = self.setUpLoggerhead()
        res = app.get("/changes", extra_environ={"REQUEST_METHOD": "HEAD"})
        self.assertEqual("text/html", res.header("Content-Type"))
        self.assertEqualDiff(b"", res.body)


def consume_app(app, env):
    body = BytesIO()
    start = []

    def start_response(status, headers, exc_info=None):
        start.append((status, headers, exc_info))
        return body.write

    extra_content = list(app(env, start_response))
    body.writelines(extra_content)
    return start[0], body.getvalue()


# class TestGlobalConfig(BasicTests):
#    """
#    Test that global config settings are respected
#    """

#    def setUp(self):
#        BasicTests.setUp(self)
#        self.createBranch()
#        config.GlobalConfig().set_user_option('http_version', 'True')

#    def test_setting_respected(self):
# FIXME: Figure out how to test this properly
#        app = self.setUpLoggerhead()
#        res = app.get('/changes', status=200)



# Copyright (C) 2011 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


from .test_simple import BasicTests


class TestRevisionUI(BasicTests):
    def test_authors_vs_committer(self):
        self.createBranch()
        self.tree.commit(
            "First",
            committer="Joe Example <joe@example.com>",
            revprops={
                "authors": "A Author <aauthor@example.com>\n"
                "B Author <bauthor@example.com>"
            },
        )
        app = self.setUpLoggerhead()
        res = app.get("/revision/1")
        # We would like to assert that Joe Example is connected to Committer,
        # and the Authors are connected. However, that requires asserting the
        # exact HTML connections, which I wanted to avoid.
        res.mustcontain("Committer", "Joe Example", "Author(s)", "A Author, B Author")

    def test_author_is_committer(self):
        self.createBranch()
        self.tree.commit("First", committer="Joe Example <joe@example.com>")
        app = self.setUpLoggerhead()
        res = app.get("/revision/1")
        # We would like to assert that Joe Example is connected to Committer,
        # and the Authors are connected. However, that requires asserting the
        # exact HTML connections, which I wanted to avoid.
        res.mustcontain("Committer", "Joe Example")
        self.assertNotIn(b"Author(s)", res.body)



#
# Copyright (C) 2006  Robey Pointer <robey@lag.net>
# Copyright (C) 2006  Goffredo Baroncelli <kreijack@inwind.it>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

import json

from breezy import urlutils
from paste.httpexceptions import HTTPServerError

from .. import util
from ..controllers import TemplatedBranchView
from ..controllers.filediff_ui import diff_chunks_for_file

DEFAULT_LINE_COUNT_LIMIT = 3000


def dq(p):
    if not isinstance(p, bytes):
        p = p.encode("UTF-8")
    return urlutils.quote(urlutils.quote_from_bytes(p, safe=""))


class RevisionUI(TemplatedBranchView):
    template_name = "revision"
    supports_json = True

    def get_values(self, path, kwargs, headers):
        h = self._history
        revid = self.get_revid()

        filter_path = kwargs.get("filter_path", None)
        start_revid = h.fix_revid(kwargs.get("start_revid", None))
        query = kwargs.get("q", None)
        compare_revid = h.fix_revid(kwargs.get("compare_revid", None))

        # TODO: This try/except looks to date before real exception handling
        # and should be removed
        try:
            revid, start_revid, revid_list = h.get_view(
                revid, start_revid, filter_path, query
            )
        except BaseException:
            self.log.exception("Exception fetching changes")
            raise HTTPServerError("Could not fetch changes")
        # XXX: Some concern about namespace collisions. These are only stored
        # here so they can be expanded into the template later. Should probably
        # be stored in a specific dict/etc.
        self.revid_list = revid_list
        self.compare_revid = compare_revid
        self.path = path
        kwargs["start_revid"] = start_revid

        change = h.get_changes([revid])[0]

        if compare_revid is None:
            file_changes = h.get_file_changes(change)
        else:
            file_changes = h.file_changes_for_revision_ids(compare_revid, change.revid)

        h.add_branch_nicks(change)

        if "." in change.revno:
            # Walk "up" though the merge-sorted graph until we find a
            # revision with merge depth 0: this is the revision that merged
            # this one to mainline.
            ri = self._history._rev_info
            i = self._history._rev_indices[change.revid]
            while ri[i][0][2] > 0:
                i -= 1
            merged_in = ri[i][0][3]
        else:
            merged_in = None

        return {
            "revid": revid.decode("utf-8"),
            "change": change,
            "file_changes": file_changes,
            "merged_in": merged_in,
        }

    def add_template_values(self, values):
        super(RevisionUI, self).add_template_values(values)
        remember = self._history.fix_revid(self.kwargs.get("remember", None))
        query = self.kwargs.get("q", None)
        filter_path = self.kwargs.get("filter_path", None)
        start_revid = self.kwargs["start_revid"]
        navigation = util.Container(
            revid_list=self.revid_list,
            revid=values["revid"],
            start_revid=start_revid,
            filter_path=filter_path,
            pagesize=1,
            scan_url="/revision",
            branch=self._branch,
            feed=True,
            history=self._history,
        )
        if query is not None:
            navigation.query = query
        util.fill_in_navigation(navigation)
        path = self.path
        if path in ("", "/"):
            path = None

        file_changes = values["file_changes"]
        link_data = {}
        path_to_id = {}
        if path:
            items = [x for x in file_changes.text_changes if x.filename == path]
            if len(items) > 0:
                item = items[0]
                try:
                    context_lines = int(self.kwargs["context"])
                except (KeyError, ValueError):
                    context_lines = None
                diff_chunks = diff_chunks_for_file(
                    self._history._branch.repository,
                    path,
                    item.old_revision,
                    item.new_revision,
                    context_lines=context_lines,
                )
            else:
                diff_chunks = None
        else:
            diff_chunks = None
            for i, item in enumerate(file_changes.text_changes):
                item.index = i
                link_data["diff-" + str(i)] = "%s/%s/%s" % (
                    dq(item.new_revision),
                    dq(item.old_revision),
                    dq(item.filename),
                )
                path_to_id[item.filename] = "diff-" + str(i)

        # Directory Breadcrumbs
        directory_breadcrumbs = util.directory_breadcrumbs(
            self._branch.friendly_name, self._branch.is_root, "changes"
        )
        can_export = self._branch.export_tarballs

        values.update(
            {
                "history": self._history,
                "link_data": json.dumps(link_data),
                "json_specific_path": json.dumps(path),
                "path_to_id": json.dumps(path_to_id),
                "directory_breadcrumbs": directory_breadcrumbs,
                "navigation": navigation,
                "remember": remember,
                "compare_revid": self.compare_revid,
                "filter_path": filter_path,
                "diff_chunks": diff_chunks,
                "query": query,
                "can_export": can_export,
                "specific_path": path,
                "start_revid": start_revid,
            }
        )



# Copyright 2006, 2010, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


from __future__ import absolute_import


def test_suite():
    import unittest

    loader = unittest.TestLoader()
    return loader.loadTestsFromNames(
        [
            (__name__ + "." + x)
            for x in [
                "test_controllers",
                "test_corners",
                "test_history",
                "test_http_head",
                "test_load_test",
                "test_simple",
                "test_revision_ui",
                "test_templating",
                "test_util",
                "test_highlight",
            ]
        ]
    )



#
# Copyright (C) 2006  Robey Pointer <robey@lag.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

"""
a cache for chewed-up 'file change' data structures, which are basically just
a different way of storing a revision delta.  the cache improves lookup times
10x over bazaar's xml revision structure, though, so currently still worth
doing.

once a revision is committed in bazaar, it never changes, so once we have
cached a change, it's good forever.
"""

import marshal
import os
import pickle
import tempfile
import zlib
from sqlite3 import dbapi2

# We take an optimistic approach to concurrency here: we might do work twice
# in the case of races, but not crash or corrupt data.


def safe_init_db(filename, init_sql):
    # To avoid races around creating the database, we create the db in
    # a temporary file and rename it into the ultimate location.
    fd, temp_path = tempfile.mkstemp(dir=os.path.dirname(filename))
    os.close(fd)
    con = dbapi2.connect(temp_path)
    cur = con.cursor()
    cur.execute(init_sql)
    con.commit()
    con.close()
    os.rename(temp_path, filename)


class FakeShelf(object):
    def __init__(self, filename):
        create_table = not os.path.exists(filename)
        if create_table:
            safe_init_db(
                filename,
                "create table RevisionData (revid binary primary key, data binary)",
            )
        self.connection = dbapi2.connect(filename)
        self.cursor = self.connection.cursor()

    def _create_table(self, filename):
        con = dbapi2.connect(filename)
        cur = con.cursor()
        cur.execute("create table RevisionData (revid binary primary key, data binary)")
        con.commit()
        con.close()

    def _serialize(self, obj):
        return dbapi2.Binary(pickle.dumps(obj, protocol=2))

    def _unserialize(self, data):
        return pickle.loads(str(data))

    def get(self, revid):
        self.cursor.execute("select data from revisiondata where revid = ?", (revid,))
        filechange = self.cursor.fetchone()
        if filechange is None:
            return None
        else:
            return self._unserialize(filechange[0])

    def add(self, revid, object):
        try:
            self.cursor.execute(
                "insert into revisiondata (revid, data) values (?, ?)",
                (revid, self._serialize(object)),
            )
            self.connection.commit()
        except dbapi2.IntegrityError:
            # If another thread or process attempted to set the same key, we
            # assume it set it to the same value and carry on with our day.
            pass


class RevInfoDiskCache(object):
    """Like `RevInfoMemoryCache` but backed in a sqlite DB."""

    def __init__(self, cache_path):
        if not os.path.exists(cache_path):
            os.mkdir(cache_path)
        filename = os.path.join(cache_path, "revinfo.sql")
        create_table = not os.path.exists(filename)
        if create_table:
            safe_init_db(
                filename,
                "create table Data (key binary primary key, revid binary, data binary)",
            )
        self.connection = dbapi2.connect(filename)
        self.cursor = self.connection.cursor()

    def get(self, key, revid):
        if not isinstance(key, bytes):
            raise TypeError(key)
        if not isinstance(revid, bytes):
            raise TypeError(revid)
        self.cursor.execute(
            "select revid, data from data where key = ?", (dbapi2.Binary(key),)
        )
        row = self.cursor.fetchone()
        if row is None:
            return None
        elif str(row[0]) != revid:
            return None
        else:
            try:
                return marshal.loads(zlib.decompress(row[1]))
            except (EOFError, ValueError, TypeError):
                return None

    def set(self, key, revid, data):
        if not isinstance(key, bytes):
            raise TypeError(key)
        if not isinstance(revid, bytes):
            raise TypeError(revid)
        try:
            self.cursor.execute("delete from data where key = ?", (dbapi2.Binary(key),))
            blob = zlib.compress(marshal.dumps(data, 2))
            self.cursor.execute(
                "insert into data (key, revid, data) values (?, ?, ?)",
                list(map(dbapi2.Binary, [key, revid, blob])),
            )
            self.connection.commit()
        except dbapi2.IntegrityError:
            # If another thread or process attempted to set the same key, we
            # don't care too much -- it's only a cache after all!
            pass



#
# Copyright (C) 2008  Canonical Ltd.
# Copyright (C) 2006  Robey Pointer <robey@lag.net>
# Copyright (C) 2006  Goffredo Baroncelli <kreijack@inwind.it>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA

import json
import time

import breezy.errors
from paste.httpexceptions import HTTPNotFound
from paste.request import parse_querystring, path_info_pop

from .. import templates, util
from ..templatefunctions import templatefunctions
from ..zptsupport import load_template


class BufferingWriter(object):
    def __init__(self, writefunc, buf_limit):
        self.bytes = 0
        self.buf = []
        self.buflen = 0
        self.writefunc = writefunc
        self.buf_limit = buf_limit

    def flush(self):
        self.writefunc(b"".join(self.buf))
        self.buf = []
        self.buflen = 0

    def write(self, data):
        self.buf.append(data)
        self.buflen += len(data)
        self.bytes += len(data)
        if self.buflen > self.buf_limit:
            self.flush()


class TemplatedBranchView(object):
    template_name = None
    supports_json = False

    def __init__(self, branch, history_callable):
        self._branch = branch
        self._history_callable = history_callable
        self.__history = None
        self.log = branch.log

    @property
    def _history(self):
        if self.__history is not None:
            return self.__history
        self.__history = self._history_callable()
        return self.__history

    def parse_args(self, environ):
        kwargs = dict(parse_querystring(environ))
        util.set_context(kwargs)
        args = []
        while True:
            arg = path_info_pop(environ)
            if arg is None:
                break
            args.append(arg)

        path = None
        if len(args) > 1:
            path = "/".join(args[1:])
            if isinstance(path, bytes):
                # Python 2
                path = path.decode("utf-8")
        self.args = args
        self.kwargs = kwargs
        return path

    def add_template_values(self, values):
        values.update(
            {
                "static_url": self._branch.static_url,
                "branch": self._branch,
                "util": util,
                "url": self._branch.context_url,
            }
        )
        values.update(templatefunctions)

    def __call__(self, environ, start_response):
        z = time.time()
        if environ.get("loggerhead.as_json") and not self.supports_json:
            raise HTTPNotFound
        path = self.parse_args(environ)
        headers = {}
        values = self.get_values(path, self.kwargs, headers)

        self.log.info(
            "Getting information for %s: %.3f secs"
            % (self.__class__.__name__, time.time() - z)
        )
        if environ.get("loggerhead.as_json"):
            headers["Content-Type"] = "application/json"
        elif "Content-Type" not in headers:
            headers["Content-Type"] = "text/html"
        writer = start_response("200 OK", list(headers.items()))
        if environ.get("REQUEST_METHOD") == "HEAD":
            # No content for a HEAD request
            return []
        z = time.time()
        w = BufferingWriter(writer, 8192)
        if environ.get("loggerhead.as_json"):
            w.write(
                json.dumps(values, default=util.convert_to_json_ready).encode("utf-8")
            )
        else:
            self.add_template_values(values)
            template = load_template("%s.%s" % (templates.__name__, self.template_name))
            template.expand_into(w, **values)
        w.flush()
        self.log.info(
            "Rendering %s: %.3f secs, %s bytes"
            % (self.__class__.__name__, time.time() - z, w.bytes)
        )
        return []

    def get_revid(self):
        h = self._history
        if h is None:
            return None
        if len(self.args) > 0 and self.args != [""]:
            try:
                revid = h.fix_revid(self.args[0])
            except breezy.errors.NoSuchRevision:
                raise HTTPNotFound
            assert isinstance(revid, bytes)
        else:
            revid = h.last_revid
        if revid is not None and not isinstance(revid, bytes):
            raise TypeError(revid)
        return revid



#
# Copyright (C) 2006  Robey Pointer <robey@lag.net>
# Copyright (C) 2006  Goffredo Baroncelli <kreijack@inwind.it>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

from ..controllers import TemplatedBranchView


class AtomUI(TemplatedBranchView):
    template_name = "atom"

    def get_values(self, path, kwargs, headers):
        history = self._history
        pagesize = int(20)  # self._branch.config.get('pagesize', '20'))

        revid_list = history.get_file_view(history.last_revid, None)
        entries = list(history.get_changes(list(revid_list)[:pagesize]))

        headers["Content-Type"] = "application/atom+xml"
        return {
            "changes": entries,
            "updated": entries[0].utc_date.isoformat(),
            "history": self._history,
        }



# Copyright (C) 2008-2011 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA

"""The WSGI application for serving a Bazaar branch."""

import logging
import sys
import wsgiref.util

import breezy.branch
import breezy.errors
import breezy.lru_cache
from breezy import urlutils
from breezy.hooks import Hooks
from paste import httpexceptions, request

from .. import util
from ..apps import health_app, static_app
from ..controllers.annotate_ui import AnnotateUI
from ..controllers.atom_ui import AtomUI
from ..controllers.changelog_ui import ChangeLogUI
from ..controllers.diff_ui import DiffUI
from ..controllers.download_ui import DownloadTarballUI, DownloadUI
from ..controllers.filediff_ui import FileDiffUI
from ..controllers.inventory_ui import InventoryUI
from ..controllers.revision_ui import RevisionUI
from ..controllers.revlog_ui import RevLogUI
from ..controllers.search_ui import SearchUI
from ..controllers.view_ui import ViewUI
from ..history import History

_DEFAULT = object()


class BranchWSGIApp(object):
    def __init__(
        self,
        branch,
        friendly_name=None,
        config={},
        graph_cache=None,
        branch_link=None,
        is_root=False,
        served_url=_DEFAULT,
        use_cdn=False,
        private=False,
        export_tarballs=True,
    ):
        """Create branch-publishing WSGI app.

        :param export_tarballs: If true, allow downloading snapshots of revisions
            as tarballs.
        """
        self.branch = branch
        self._config = config
        self.friendly_name = friendly_name
        self.branch_link = branch_link  # Currently only used in Launchpad
        self.log = logging.getLogger("loggerhead.%s" % (friendly_name,))
        if graph_cache is None:
            graph_cache = breezy.lru_cache.LRUCache(10)
        self.graph_cache = graph_cache
        self.is_root = is_root
        self.served_url = served_url
        self.use_cdn = use_cdn
        self.private = private
        self.export_tarballs = export_tarballs

    def public_private_css(self):
        if self.private:
            return "private"
        else:
            return "public"

    def get_history(self):
        revinfo_disk_cache = None
        cache_path = self._config.get("cachepath", None)
        if cache_path is not None:
            # Only import the cache if we're going to use it.
            # This makes sqlite optional
            try:
                from ..changecache import RevInfoDiskCache
            except ImportError:
                self.log.debug(
                    "Couldn't load python-sqlite, continuing without using a cache"
                )
            else:
                revinfo_disk_cache = RevInfoDiskCache(cache_path)
        return History(
            self.branch,
            self.graph_cache,
            revinfo_disk_cache=revinfo_disk_cache,
            cache_key=(
                self.friendly_name.encode("utf-8") if self.friendly_name else None
            ),
        )

    # Before the addition of this method, clicking to sort by date from
    # within a branch caused a jump up to the top of that branch.
    def sort_url(self, *args, **kw):
        if isinstance(args[0], list):
            args = args[0]
        qs = []
        for k, v in kw.items():
            if v is not None:
                qs.append("%s=%s" % (k, urlutils.quote(v)))
        qs = "&".join(qs)
        path_info = self._path_info.strip("/").split("?")[0]
        path_info += "?" + qs
        return self._url_base + "/" + path_info

    def url(self, *args, **kw):
        if isinstance(args[0], list):
            args = args[0]
        qs = []
        for k, v in kw.items():
            if v is not None:
                qs.append("%s=%s" % (k, urlutils.quote(v)))
        qs = "&".join(qs)
        path_info = urlutils.quote("/".join(args), safe="/~:")
        if qs:
            path_info += "?" + qs
        return self._url_base + path_info

    def absolute_url(self, *args, **kw):
        rel_url = self.url(*args, **kw)
        return request.resolve_relative_url(rel_url, self._environ)

    def context_url(self, *args, **kw):
        kw = util.get_context(**kw)
        return self.url(*args, **kw)

    def static_url(self, path):
        return self._static_url_base + path

    def js_library_url(self, path):
        if self.use_cdn:
            if path == "jquery.min.js":
                return (
                    "https://ajax.googleapis.com/ajax/libs/jquery/3.2.1/jquery.min.js"
                )
            raise KeyError("unknown js library %s" % path)
        else:
            return self.static_url("/static/javascript/" + path)

    controllers_dict = {
        "+filediff": FileDiffUI,
        "+revlog": RevLogUI,
        "annotate": AnnotateUI,
        "atom": AtomUI,
        "changes": ChangeLogUI,
        "diff": DiffUI,
        "download": DownloadUI,
        "files": InventoryUI,
        "revision": RevisionUI,
        "search": SearchUI,
        "view": ViewUI,
        "tarball": DownloadTarballUI,
    }

    def last_updated(self):
        h = self.get_history()
        change = h.get_changes([h.last_revid])[0]
        return change.date

    def public_branch_url(self):
        return self.branch.get_public_branch()

    def lookup_app(self, environ):
        # Check again if the branch is blocked from being served, this is
        # mostly for tests. It's already checked in apps/transport.py
        if not self.branch.get_config().get_user_option_as_bool(
            "http_serve", default=True
        ):
            raise httpexceptions.HTTPNotFound()
        self._url_base = environ["SCRIPT_NAME"]
        self._path_info = environ["PATH_INFO"]
        self._static_url_base = environ.get("loggerhead.static.url")
        if self._static_url_base is None:
            self._static_url_base = self._url_base
        self._environ = environ
        if self.served_url is _DEFAULT:
            public_branch = self.public_branch_url()
            if public_branch is not None:
                self.served_url = public_branch
            else:
                self.served_url = wsgiref.util.application_uri(environ)
        for hook in self.hooks["controller"]:
            controller = hook(self, environ)
            if controller is not None:
                return controller
        path = request.path_info_pop(environ)
        if not path:
            raise httpexceptions.HTTPMovedPermanently(self.absolute_url("/changes"))
        if path == "health":
            return health_app
        if path == "static":
            return static_app
        elif path == "+json":
            environ["loggerhead.as_json"] = True
            path = request.path_info_pop(environ)
        cls = self.controllers_dict.get(path)
        if cls is not None:
            return cls(self, self.get_history)
        raise httpexceptions.HTTPNotFound()

    def app(self, environ, start_response):
        with self.branch.lock_read():
            try:
                c = self.lookup_app(environ)
                return c(environ, start_response)
            except:
                environ["exc_info"] = sys.exc_info()
                environ["branch"] = self
                raise


class BranchWSGIAppHooks(Hooks):
    """A dictionary mapping hook name to a list of callables for WSGI app branch hooks."""

    def __init__(self):
        """Create the default hooks."""
        Hooks.__init__(
            self, "breezy.plugins.loggerhead.apps.branch", "BranchWSGIApp.hooks"
        )
        self.add_hook(
            "controller",
            "Invoked when looking for the controller to use for a "
            "branch subpage. The api signature is (branch_app, environ)."
            "If a hook can provide a controller, it should return one, "
            "as a standard WSGI app. If it can't provide a controller, "
            "it should return None",
            (1, 19),
        )


BranchWSGIApp.hooks = BranchWSGIAppHooks()



# daemon code from ASPN
#

import os


def daemonize(pidfile, home):
    """
    Detach this process from the controlling terminal and run it in the
    background as a daemon.
    """

    WORKDIR = "/"
    REDIRECT_TO = getattr(os, "devnull", "/dev/null")
    MAXFD = 1024

    try:
        pid = os.fork()
    except OSError as e:
        raise Exception("%s [%d]" % (e.strerror, e.errno))

    if pid == 0:  # The first child.
        os.setsid()

        try:
            pid = os.fork()  # Fork a second child.
        except OSError as e:
            raise Exception("%s [%d]" % (e.strerror, e.errno))

        if pid == 0:  # The second child.
            os.chdir(WORKDIR)
        else:
            os._exit(0)  # Exit parent (the first child) of the second child.
    else:
        os._exit(0)  # Exit parent of the first child.

    import resource  # Resource usage information.

    maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
    if maxfd == resource.RLIM_INFINITY:
        maxfd = MAXFD

    fd = os.open(REDIRECT_TO, os.O_RDONLY)
    os.dup2(fd, 0)
    fd = os.open(REDIRECT_TO, os.O_WRONLY)
    os.dup2(fd, 1)
    os.dup2(fd, 2)

    # Iterate through and close all other file descriptors.
    for fd in range(3, maxfd):
        try:
            os.close(fd)
        except OSError:  # ERROR, fd wasn't open to begin with (ignored)
            pass

    f = open(pidfile, "w")
    f.write("%d\n" % (os.getpid(),))
    f.close()


def is_running(pidfile):
    try:
        f = open(pidfile, "r")
    except IOError:
        return False
    pid = int(f.readline())
    f.close()
    try:
        os.kill(pid, 0)
    except OSError:
        # no such process
        return False
    return True



# Copyright (C) 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from .test_simple import TestWithSimpleTree


class TestInventoryUI(TestWithSimpleTree):
    def test_authors_vs_committer(self):
        app = self.setUpLoggerhead()
        res = app.get("/files")
        # download url in top directory is composed correctly
        res.mustcontain("/download/rev-1/myfilename")

        res2 = app.get("/files/head:/folder")
        # download url in subdirectory is composed correctly
        res2.mustcontain("/download/rev-1/folder/myfilename")



#
# Copyright (C) 2008, 2009 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA

"""A simple container to turn this into a python package."""

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata

try:
    __version__ = importlib_metadata.version("loggerhead")
except importlib_metadata.PackageNotFoundError:
    # Support running tests from the build tree without installation.
    import os

    try:
        from setuptools.config.setupcfg import read_configuration
    except ModuleNotFoundError:
        from setuptools.config import read_configuration
    cfg = read_configuration(os.path.join(os.path.dirname(__file__), "..", "setup.cfg"))
    __version__ = cfg["metadata"]["version"]
__revision__ = None
required_breezy = (3, 1)
